//
//      Hugo KUENY - nessotrin 2015
//      Version 1.3.0


#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cmath>

#include "ComputedPacket.h"
#include "Packet.h"
#include "Tools.h"
#include "IO.h"
#include "File.h"
#include "time.c"


#include <libusb-1.0/libusb.h>

using namespace std;

int connectToDevice();

libusb_context *context = NULL;
libusb_device **deviceList = NULL;


void DumperHelp(int argCount)
{
    printf("Error: file name requierd, %d arguments give given\n",argCount);
    printf("Example: DUMPER memoryDump.dmp\n");
}

void LoggerHelp(int argCount)
{
    printf("Error: no arguments requierd, %d given\n",argCount);
}

int LoggerMode()
{
    printf("Logging: use CTRL+C to quit !\n");
    while(1)
    {
        while(connected == 0)
        {
            connected = connectToDevice();
        }

        char character[512];
        memset(character,'\0',512);
        if(Get512Character(character) > 0)
        {
            printf("%s",character);
        }


    }
}

#define DUMPER_FILL_TIMEOUT 1000
#define DUMPER_SEND_SIZE 0x40
#define DUMPER_EFFECTIVE_SEND_SIZE (DUMPER_SEND_SIZE*2+2)
#define DUMPER_TO_SEND (0xffff+1) //65536 et pas 65535
#define DUMPER_PACKET_COUNT (DUMPER_TO_SEND/DUMPER_SEND_SIZE)

int FillDumpBuffer(unsigned char * queueBuffer, int * bufferCount)
{
    int startTime = getTime();

    while(*bufferCount < DUMPER_EFFECTIVE_SEND_SIZE) // on attend un packet
    {
        *bufferCount += Get512Character((char *)queueBuffer+*bufferCount);
        long endTime = getTime();
        if((int)(endTime - startTime) > DUMPER_FILL_TIMEOUT)
        {
            printf("End after %dms without data\n",(int)(endTime - startTime));
            return 1;
        }
    }
    return 0;
}

int DumperMode(int argc, char *argv[])
{
    char * filename = argv[2];

    #ifndef RELEASE
    if(argc != 3)
    {
        filename = (char*)"dump.dat";
    }
    #endif // RELEASE

    int sectorCount = 0;

    unsigned char dataBuffer[DUMPER_TO_SEND];
    for(int i = 0 ; i < DUMPER_TO_SEND ; i++)
        dataBuffer[i] = 0;

    unsigned char queueBuffer[4096000];

    unsigned char completionMap[DUMPER_PACKET_COUNT];

    for(int i = 0 ; i < DUMPER_PACKET_COUNT ; i++)
        completionMap[i] = 0;

    int bufferCount = 0;

    while(connected == 0)
    {
        connected = connectToDevice();
    }

    printf("Beginning...\n");

    int byteCount = 0;

    unsigned int total_checksum = 0;

    while(1) //sectorCount < TO_SEND/SEND_SIZE
    {

        if(FillDumpBuffer(queueBuffer,&bufferCount))
            break;

        int position = ((queueBuffer[0]&0xff) + ((queueBuffer[1]&0xff)<<8) -1)*DUMPER_SEND_SIZE;
        int positionId = ((queueBuffer[0]&0xff) + ((queueBuffer[1]&0xff)<<8) -1);

        if((queueBuffer[0]&0xff) + ((queueBuffer[1]&0xff)<<8) == 0) //all sector id are offset by 1
        {
            printf("Blank or currupted data received, stop the dump and retry (start the dumper before the addin !!) !\n");
            exit(1);
        }

        completionMap[positionId]++;

        if(completionMap[positionId] > 1)
        {
            printf("BAD: SECTOR OVERLAP, sector %d, count %d, transfere count %d\nData is probably corrupted !\n",positionId,completionMap[positionId],sectorCount);
        }

        //printf("New part: (%d/%d), ID:%d\n",sectorCount+1,TO_SEND/SEND_SIZE,position/SEND_SIZE);

        unsigned char last = 0;

        for(int i = 0 ; i < DUMPER_SEND_SIZE ; i++)
        {

            unsigned char value = queueBuffer[i*2+2];
            unsigned char checksum = queueBuffer[i*2+2+1];
            unsigned char next_value = queueBuffer[(i+1)*2+2];
            unsigned char next_checksum = queueBuffer[(i+1)*2+2+1];

            if((unsigned char)((unsigned char)last+(unsigned char)value) != (unsigned char)checksum)
            {
                if((unsigned char)((unsigned char)value+(unsigned char)next_value) != (unsigned char)next_checksum)
                {
                    short temp = checksum-last;
                    if(checksum < value || checksum < last) // overflow à l'additon
                    {
                        temp += 256;
                    }
                    printf("Patched: %d was %d\n",(unsigned char)value, (unsigned char)temp&0xff);

                    value = temp & 0xff;
                }
                else
                {
                }

            }

            total_checksum += ( dataBuffer[position+i] = queueBuffer[i*2+2] );

            last = value;
            byteCount++;
        }


        memcpy(queueBuffer+DUMPER_EFFECTIVE_SEND_SIZE,queueBuffer,bufferCount-DUMPER_EFFECTIVE_SEND_SIZE); // on retire
        bufferCount -= DUMPER_EFFECTIVE_SEND_SIZE;

        sectorCount++;

    }

    FILE * dumpFile;

    dumpFile = fopen(filename,"w");
    if(dumpFile == NULL)
    {
        printf("FILe: Can't open dump file !\n");
        return 1;
    }

    if(fwrite(dataBuffer , sizeof(char), DUMPER_TO_SEND, dumpFile) != DUMPER_TO_SEND)
    {
        printf("FILE: Unknown write error !\n");
        return 1;
    }

    fclose(dumpFile);



    printf("Done !\nChecksum: %d\n",total_checksum);

    return 0;
}

void SendfileHelp()
{
    printf("Error: please use these arguments:\n");
    printf("Format:  file_to_send       file_name  file_folder  file_device\n");
    printf("Example: MyFiles/addin.g1a  addin.g1a  game         fls0\n");
    printf("Example: ../Game/help.txt   data.txt                crd0\n");
}

int SendfileMode(int argc, char *argv[])
{
    char * filenameToSend = argv[2];
    char * filename = argv[3];
    char * filefolder = argv[4];
    char * filedevice = argv[5];


    #ifndef RELEASE
    if(argc != 5 && argc != 6) // debug only
    {
        printf((char*)"Dev mode activate, dev values selected !\n");
        filenameToSend = (char*)"clock.g1a";
        filename = (char*)"clock.g1a";
        filefolder = (char*)"";
        filedevice = (char*)"fls0";
    }
    #endif

    if(argc == 5) // no folder indicated, reordering arguments
    {
        filedevice = filefolder;
        filefolder = (char*)"";
    }


    FILE * fileToSend;

    fileToSend = fopen(filenameToSend,"rb");
    if(fileToSend == NULL)
    {
        printf("ERROR: Can't open file to transfere ! Did you misspelled the file ?\n");
        return 1;
    }

    if(fseek(fileToSend, 0, SEEK_END))
    {
        printf("ERROR: can't seek to the end of the file !\n");
        return 1;
    }
    unsigned int fileSize = ftell(fileToSend);
    if(fseek(fileToSend, 0, SEEK_SET))
    {
        printf("ERROR: can't seek to the start of the file !\n");
        return 1;
    }

    if(fileSize == 0)
    {
        printf("ERROR: file seems empty !\n");
        return 1;
    }

    char * buffer = (char *) malloc(fileSize);
    if(buffer == NULL)
    {
        printf("ERROR: Out of memory for the file buffer !\n");
        return 1;
    }


    if(fread (buffer,1,fileSize,fileToSend) != fileSize)
    {
        printf("ERROR: Unknown error while reading the file !\n");
        return 1;
    }

    fclose(fileToSend);


    printf("Sendfile: File size: %do\n",fileSize);

    SendFile(filename,fileSize,buffer,filefolder,filedevice,0x02);

    return 0;
}


void modeHelp()
{
    printf("Please use one of the following modes:\n");
    printf("SEND\n");
    printf("LOGGER\n");
    printf("DUMPER");
    printf("\n");
    printf("Exemple:\n");
    printf("UsbLogger SEND MyAddinName.G1A\n addin.g1a gameFold fls0\n");
    printf("UsbLogger DUMPER dump.dat\n");
    printf("UsbLogger LOGGER\n");
}

int connectToDevice()
{
    unsigned int deviceCount = libusb_get_device_list(context, &deviceList);

    libusb_device * device = NULL;

    for (size_t idx = 0; idx < deviceCount; ++idx)
    {
        libusb_device *tmpDevice = deviceList[idx];
        libusb_device_descriptor deviceDescriptor = {0};

        if(libusb_get_device_descriptor(tmpDevice, &deviceDescriptor) != 0)
        {
            printf("WARNING: Can't get a device's usb descriptor !");
        }

        if(deviceDescriptor.idVendor == 0x07CF && deviceDescriptor.idProduct == 0x6101)
        {
            device = tmpDevice;
        }

        //printf("Vendor:Device = %04x:%04x\n", desc.idVendor, desc.idProduct);
    }

    if(device == NULL)
    {
        return 0;
    }


    int result = libusb_open(device, &deviceHandle);

    if(result != 0)
    {
        if(result == LIBUSB_ERROR_ACCESS)
        {
            printf("ERROR: Not enought permission to open the calculator !\n");
            printf("Run as root or add a rule to udev\n");
            return -1;
        }
        else
        {
            printf("ERROR: Unknown error while opening the calculator !");
            return -1;
        }
    }

    connected = 1;

    return 1;

}

int main(int argc, char *argv[])
{
    printf("UsbConnector 1.3.0 [IN ACTIVE DEVELOPPEMENT]\n");

    char * modeName = argv[1];

    if(argc < 2)
    {
        #ifdef RELEASE
        modeHelp();
        return 1;
        #else
        printf("Devmode & no arguments => auto set to LOGGER\n");
        modeName = (char*)"LOGGER";
        #endif // RELEASE
    }

    int modeId = -1;

    if(memcmp(modeName,"SEND",4) == 0)
    {
        if(argc == 5 or argc == 6)
        {
            modeId = 0;
        }
        else
        {
            #ifdef RELEASE
            SendfileHelp();
            #else
            modeId = 0; // default mode for DEBUG
            #endif
        }
    }
    else if(memcmp(modeName,"LOGGER",6) == 0)
    {
        if(argc == 2)
        {
            modeId = 1;
        }
        else
        {
            #ifdef RELEASE
            LoggerHelp(argc);
            #else
            modeId = 1;
            #endif
        }
    }
    else if(memcmp(modeName,"DUMPER",6) == 0)
    {
        if(argc == 3)
        {
            modeId = 2;
        }
        else
        {
            #ifdef RELEASE
            DumperHelp(argc);
            #else
            modeId = 2;
            #endif
        }
    }


    if(modeId == -1)
    {
        modeHelp();
        return 1;
    }



    if(libusb_init(&context) != 0)
    {
        printf("ERROR: Can't initialize libusb");
        return 1;
    }




    if(modeId == 0)
    {
        int connectResult = connectToDevice();
        if(connectResult == 1)
        {
            printf("Connected to the calculator !\n");
        }
        else
        {
            if(connectResult == 0)
            {
                printf((char *)"ERROR: Can't acces the calculator !\n");
                printf("Activate receive mode ?\n");
            }
            return 1;
        }

        InitConnection();
        printf("File mode selected !\n");
        if(SendfileMode(argc, argv))
        {
            return 1;
        }
        EndConnection();
    }
    if(modeId == 1)
    {
        printf("Logger mode selected !\n");
        if(LoggerMode())
        {
            return 1;
        }

    }
    if(modeId == 2)
    {
        printf("Dumper mode selected ! \n");
        if(DumperMode(argc, argv))
        {
            return 1;
        }

    }


    libusb_free_device_list(deviceList, 1);

}

